#
import datetime
import hashlib
import hmac
import requests
from requests.utils import quote

def generate_aws_signed_url(region, bucket, object_key, access_key, secret_key, expiration, endpoint=None):

    # request elements
    http_method = 'GET'
    endpoint = endpoint if endpoint else 's3.amazonaws.com'
    host = bucket + '.' + endpoint
    endpoint = 'https://' + host

    # hashing methods
    def hash(key, msg):
        return hmac.new(key, msg.encode('utf-8'), hashlib.sha256).digest()

    # region is a wildcard value that takes the place of the AWS region value
    # as COS doen't use regions like AWS, this parameter can accept any string
    def createSignatureKey(key, datestamp, region, service):
        keyDate = hash(('AWS4' + key).encode('utf-8'), datestamp)
        keyRegion = hash(keyDate, region)
        keyService = hash(keyRegion, service)
        keySigning = hash(keyService, 'aws4_request')
        return keySigning

    # assemble the standardized request
    time = datetime.datetime.utcnow()
    timestamp = time.strftime('%Y%m%dT%H%M%SZ')
    datestamp = time.strftime('%Y%m%d')

    standardized_querystring = ('X-Amz-Algorithm=AWS4-HMAC-SHA256' +
                                '&X-Amz-Credential=' + access_key + '/' + datestamp + '/' + region + '/s3/aws4_request' +
                                '&X-Amz-Date=' + timestamp +
                                '&X-Amz-Expires=' + str(expiration) +
                                '&X-Amz-SignedHeaders=host')
    standardized_querystring_url_encoded = quote(standardized_querystring, safe='&=')

    standardized_resource = '/' + object_key
    standardized_resource_url_encoded = quote(standardized_resource, safe='&')

    payload_hash = 'UNSIGNED-PAYLOAD'
    standardized_headers = 'host:' + host
    signed_headers = 'host'

    standardized_request = (http_method + '\n' +
                            standardized_resource + '\n' +
                            standardized_querystring_url_encoded + '\n' +
                            standardized_headers + '\n' +
                            '\n' +
                            signed_headers + '\n' +
                            payload_hash).encode('utf-8')

    # assemble string-to-sign
    hashing_algorithm = 'AWS4-HMAC-SHA256'
    credential_scope = datestamp + '/' + region + '/' + 's3' + '/' + 'aws4_request'
    sts = (hashing_algorithm + '\n' +
           timestamp + '\n' +
           credential_scope + '\n' +
           hashlib.sha256(standardized_request).hexdigest())

    # generate the signature
    signature_key = createSignatureKey(secret_key, datestamp, region, 's3')
    signature = hmac.new(signature_key,
                         (sts).encode('utf-8'),
                         hashlib.sha256).hexdigest()

    # create and send the request
    # the 'requests' package autmatically adds the required 'host' header
    request_url = (endpoint + '/' +
                   object_key + '?' +
                   standardized_querystring_url_encoded +
                   '&X-Amz-Signature=' +
                   signature)

    def hex_hash(key, msg):
        return hmac.new(b'key', msg.encode('utf-8'), hashlib.sha256).hexdigest()

    def createHexSignatureKey(key, datestamp, region, service):
        keyDate = hex_hash(('AWS4' + key).encode('utf-8'), datestamp)
        keyRegion = hex_hash(keyDate, region)
        keyService = hex_hash(keyRegion, service)
        keySigning = hex_hash(keyService, 'aws4_request')
        return keySigning

    signature_key_hex = createHexSignatureKey(secret_key, datestamp, region, 's3')

    # print(request_url)
    return request_url


import sys
import logging
import json

def is_number(s):
  try:
      float(s)
      return True
  except:
      return False

def get_percent(p):
  if is_number(p):
    return round(p)
  return 0

def is_json(value):
  try:
    json.loads(value)
  except ValueError:
    return False
  return True

def rm_attributes_if_null(class_obj):
  for key, val in class_obj.__dict__.copy().items():
    if not val:
      class_obj.__dict__.pop(key)

def rm_key_if_null(obj):
  if isinstance(obj, dict):
    return _rm_key(obj)
  elif isinstance(obj, (str, bytes)):
    res = _rm_key(json.loads(obj))
    return json.dumps(res)

def _rm_key(_dict):
  for key, val in _dict.copy().items():
    if not val:
      _dict.pop(key)
  return _dict


def progress_bar(self, custom_message=None):
  message = custom_message if custom_message else ''
  while 1:
    barLength, status = 20, ""
    progress = float(self.percent) / 100.0
    if progress >= 1.:
      progress, status = 1, "\r\n"
    block = int(round(barLength * progress))
    text = "\r{} [{}] {:.0f}% {}".format(message,
             "#" * block + "-" * (barLength - block), round(progress * 100, 0), status)
    sys.stdout.write(text)
    sys.stdout.flush()
    if self.task_completed:
      break

def log(self, path=None, name=None, log_format=None):
  format = '[%(asctime)s] %(levelname)s  %(message)s' if not log_format else log_format
  name = name if name else '{0}.log'.format(self.task.token)
  path = path if path else ''
  log_name = '{0}{1}'.format(path, name)
  logging.basicConfig(filename=log_name, format=format())
  logging.getLogger().setLevel(logging.INFO)
  log = logging.getLogger()
  while 1:
   log.info('{0} | {1} | {2}'.format(self.status, self.percent, self.message))
   if self.task_completed:
     break



def client(api_key, api_url=None, version=None, **kwargs):
    from . client import Client
    return Client(api_key, api_url=api_url, version=version, **kwargs)

def custom_params():
  from . custom_params import CustomTranscodingParams
  return CustomTranscodingParams()

def format():
  from . custom_params import Format
  return Format()

def destination():
  from . custom_params import Destination
  return Destination()

def stream():
  from . custom_params import Stream
  return Stream()

def x264_video_codec():
  from . custom_params import Libx264_VideoCodecParameters
  return Libx264_VideoCodecParameters()

def x265_video_codec():
  from . custom_params import Libx265_VideoCodecParameters
  return Libx265_VideoCodecParameters()

from . exeptions import QencodeClientException, QencodeTaskException

from . tools import generate_aws_signed_url



__version__ = "0.9.30"
__status__ = "Beta"
__author__ = "Qencode"






class QencodeException(Exception):
  def __init__(self, message,  *args):
    super(QencodeException, self).__init__(message, *args)
    self.error = message
    self.arg = [i for i in args]

class QencodeClientException(QencodeException):
  def __init__(self, message,  *args):
    super(QencodeClientException, self).__init__(message, *args)


class QencodeTaskException(QencodeException):
  def __init__(self, message,  *args):
    super(QencodeTaskException, self).__init__(message, *args)


REPEAT = 32
SLEEP_REGULAR = 10
SLEEP_ERROR = 60
COMPLETED_STATUS = ['completed', 'saved']

ERROR_OK                = 0
ERROR_SERVER_INTERNAL   = 1
ERROR_BAD_APP_ID        = 2
ERROR_APP_ID_NOT_FOUND  = 3
ERROR_BAD_TOKEN         = 4
ERROR_TOKEN_NOT_FOUND   = 5
ERROR_TARIFF_NOT_PAID   = 6
ERROR_MASTER_NOT_FOUND  = 7
ERROR_SYSTEM_BUSY       = 8
ERROR_BAD_PAYLOAD       = 9
ERROR_PROJECT_NOT_FOUND = 10
ERROR_BAD_PROFILE       = 11
ERROR_PROFILE_NOT_FOUND = 12
ERROR_BAD_TOKENS        = 13
ERROR_FIELD_REQUIRED    = 14


from . httptools import Http
from . task import Task
from . metadata import Metadata

class Client(object):

  """
      :return: encoder object

  """
  def __init__(self, api_key, api_url=None, version=None):
    self.api_key = api_key
    self.api_url = api_url if api_url else 'https://api.qencode.com/'
    self.version = version if version else 'v1'
    self.connect = Http(self.version, self.api_url)
    self.access_token = None
    self.expire = None
    self.error = None
    self.code = None
    self.message = ''
    self._get_access_token()

  def create_task(self, **kwargs):
    return Task(self.access_token, self.connect, **kwargs)

  def refresh_access_token(self, **kwargs):
    response = self.connect.request('access_token', dict(api_key=self.api_key))
    if not response['error']:
      self.access_token = response['token']
      self.expire = response['expire']
    else:
      self.error = response['error']
      self.code = response['error']
      self.message = response.get('message')

  def _get_access_token(self):
    response = self.connect.request('access_token', dict(api_key=self.api_key))
    if not response['error']:
      self.access_token = response['token']
      self.expire = response['expire']
    else:
     self.error = response['error']
     self.code = response['error']
     self.message = response.get('message')

  def get_metadata(self, uri):
    metadata = Metadata(self.access_token, self.connect)
    video_info = metadata.get(uri)
    return video_info



import json
# import urllib
import urllib.request as urllib2
from urllib.parse import urljoin
from urllib.parse import urlencode

class Http(object):
  def __init__(self, version, url, debug=False):
    self.version = version
    self.url = url
    self._debug = debug

  def _call_server(self, url, post_data):
    if not url:
      response = dict(error=True, message='AttributeError: Bad URL: ' + str(url))
      return json.dumps(response)
    data = urlencode(post_data).encode("utf-8")
    request = urllib2.Request(url, data)
    #print('call_server: ', url, data)
    try:
      res = urllib2.urlopen(request)
    except urllib2.HTTPError as e:
      headers = e.headers if self._debug else ''
      response = dict(error=True, message='HTTPError: {0} {1} {2}'.format(e.code, e.reason, headers))
      response = json.dumps(response)
    except urllib2.URLError as e:
      response = dict(error=True, message='URLError: {0}'.format(e.reason))
      response = json.dumps(response)
    except BaseException as e:
      response = dict(error=True, message='Error: {0}'.format(e))
      response = json.dumps(response)
    else:
      res2 = res.read()
      response = res2 if isinstance(res2, str) else res2.decode('utf-8')      
    return response

  def request(self, api_name, data):
    path = '{version}/{api_name}'.format(version=self.version, api_name=api_name)
    response = self._call_server(urljoin(self.url, path), data)
    try:
      response = json.loads(response)
    except ValueError as e:
      response = dict(error=True, message=repr(e))
    return response

  def post(self, url, data):
    response = self._call_server(url, data)
    try:
      response = json.loads(response)
    except ValueError as e:
      response = dict(error=True, message=repr(e))
    return response



from . custom_params import Query, CustomTranscodingParams
from . const import *
import time
import json
from . utils import is_json, rm_key_if_null



class Task(object):
  def __init__(self, access_token, connect, debug=False, **kwargs):
    self.connect = connect
    self.status_url = None
    self.main_status_url = '{0}/{1}/status'.format(self.connect.url, self.connect.version)
    self.task_token = None
    self.upload_url = None
    self.access_token = access_token
    self._debug = debug
    self.message = ''
    self.error = None
    self.repeat = kwargs.get('repeats') if kwargs.get('repeats') else REPEAT
    self._create_task(1)



  def start(self, profiles, video_url, **kwargs):
    """Creating task and starting encode

      :param profiles: String or List object. Profile uuid
      :param transfer_method: String. Transfer method uuid
      :param video_url: String. Url of source video
      :param payload: String.
      :return: None

    """
    if not self.error:
      #self._create_task(1)
      data = self._prepare_data(profiles, video_url, **kwargs)

      if not self.error and self.task_token:
        self._start_encode('start_encode', data)


  def custom_start(self, data,  **kwargs):
    """Creating task and starting encode

          :param query: JSON object for query param. For examples: https://docs.qencode.com
          :param payload: String.
          :return: None

    """
    if data is None:
      self.error = True
      self.message = 'Params is required'
      
    #if not self.error:
    #  self._create_task(1)

    if not self.error:
      query = self._prepare_query(data)

      if not self.error:
        data = self._prepare_data_custom(query, **kwargs)

    if not self.error and self.task_token:
      self._start_encode('start_encode2', data)


  def status(self):
    return self._status()

  def main_status(self):
    return self._status2()

  def progress_changed(self, callback, *args, **kwargs):
    while 1:
      status = self._status()
      if status['error']:
        return callback(status, *args, **kwargs)
      callback(status, *args, **kwargs)
      if status.get('status') in COMPLETED_STATUS:
        break
      time.sleep(SLEEP_REGULAR)


  def task_completed(self, callback, *args, **kwargs):
    while 1:
      status = self._status()
      if status['error']:
        return callback(status, *args, **kwargs)
      if status.get('status') in COMPLETED_STATUS:
        return callback(status, *args, **kwargs)
      if status.get('status') in COMPLETED_STATUS:
        break
      time.sleep(SLEEP_REGULAR)


  def _prepare_data(self, profiles, video_url, **kwargs):
    data = dict(
      task_token=self.task_token,
      profiles=', '.join(profiles) if type(profiles).__name__ == 'list' else profiles
    )
    if isinstance(video_url, list):
      try:
        data.update(stitch=json.dumps(video_url))
      except Exception:
        data.update(stitch=video_url)
    else:
      data.update(uri=video_url)
    if kwargs:
      data.update(kwargs)
    return data

  def _prepare_query(self, params):    
    if isinstance(params, CustomTranscodingParams):
      query_obj = Query()
      query_obj.params = params
      query_obj.validate_params()
      if query_obj.error:
        self.error = query_obj.error
        self.message = query_obj.message
      query_obj.prepare_params()
      if query_obj.error:
        self.error = query_obj.error
        self.message = query_obj.message
      return query_obj.query
    
    if isinstance(params, dict):
      query = rm_key_if_null(params)      
      return json.dumps(query)
    
    if isinstance(params, (str, bytes)):
      if is_json(params):
        query = rm_key_if_null(params)
        return query
      else:
        self.error = True
        try:
          self.message = "JSON is not well formatted: {0} Is not defined".format(params)
        except Exception as e:
          pass
        finally:
          self.message = "JSON is not well formatted"

  def _prepare_data_custom(self, query_json, **kwargs):
    data = dict(
      task_token=self.task_token,
      query=query_json
    )
    if kwargs:
      data.update(kwargs)
    return data


  def _create_task(self, count):
    res = self.connect.request('create_task', dict(token=self.access_token))
    if not res['error']:
      self.task_token = res.get('task_token')
      self.upload_url = res.get('upload_url')
    else:
      self.error = res['error']
      self.message = res.get('message')

    if self.error and self.error == 8:
      if count < REPEAT:
        time.sleep(SLEEP_ERROR)
        self._create_task(count + 1)

  def _start_encode(self, api_name, data):
    res = self.connect.request(api_name, data)
    if not res['error'] and res.get('status_url'):
      self.status_url = res['status_url']
    else:
      self.status_url = self.main_status_url
      self.error = res.get('error')
      self.message = res.get('message')

  def _status(self):
    response = self.connect.post(self.status_url, dict(task_tokens=self.task_token))
    status = None

    if response['error'] == ERROR_BAD_TOKENS:
      raise ValueError('Bad token: ' + str(self.task_token))

    if 'statuses' in response and self.task_token in response['statuses']:
      status = response['statuses'][self.task_token]

    if not status and self.status_url != self.main_status_url:
      self.status_url = self.main_status_url
      response = self.connect.post(self.status_url, dict(task_tokens=self.task_token))
      if 'statuses' in response and self.task_token in response['statuses']:
        status = response['statuses'][self.task_token]

    if status and 'status_url' in status:
      self.status_url = status['status_url']

    return status


  # def _status2(self):
  #   response = self.connect.request('status', {'task_tokens[]': self.task_token})
  #   if not response['error']:
  #     res = response['statuses'][self.task_token]
  #     if res:
  #       return res
  #     else:
  #       return dict(error=True, message='Error getting status')
  #   else:
  #     return response




import json
from json import JSONEncoder
from . utils import rm_attributes_if_null

class CustomTranscodingParams(object):
  """CustomTranscodingParams

    :var source: String. Source video URI. Can be http(s) url or tus uri
    :var format: String. A list of objects, each describing params for a single output video stream (MP4, WEBM, HLS or MPEG-DASH)

  """
  def __init__(self):
    self.source = None
    self.format = None
    self.callback_url = None
    rm_attributes_if_null(self)

  def remove_null_params(self):
    rm_attributes_if_null(self)

class Format(object):
  """
      :var
      :var

    """
  def __init__(self):
    self.output = None
    self.file_extension = None
    self.destination = None
    self.segment_duration = None
    self.stream = None
    self.logo = None
    self.start_time = None
    self.duration = None
    self.is_watermark = None
    self.size = None
    self.video_codec = None
    self.audio_codec = None
    self.aspect_ratio = None
    self.quality = None
    self.interval = None
    self.width = None
    self.height = None
    self.time = None
    self.path = None
    self.resize_mod = None
    rm_attributes_if_null(self)

  def remove_null_params(self):
    rm_attributes_if_null(self)

class Destination(object):
  def __init__(self):
    self.url = None
    self.key = None
    self.secret = None
    self.permissions = None
    self.storage_class = None
    rm_attributes_if_null(self)

  def remove_null_params(self):
    rm_attributes_if_null(self)

class Stream(object):
  def __init__(self):
    self.size = None
    self.video_codec = None
    self.bitrate = None
    self.quality = None
    self.rotate = None
    self.framerate = None
    self.pix_format = None
    self.profile = None
    self.video_codec_parameters = None
    self.keyframe = None
    self.segment_duration = None
    self.start_time = None
    self.duration = None
    self.audio_bitrate = None
    self.audio_sample_rate = None
    self.audio_channels_number = None
    self.audio_codec = None
    self.downmix_mode = None
    self.logo = None
    self.aspect_ratio = None
    rm_attributes_if_null(self)

  def remove_null_params(self):
    rm_attributes_if_null(self)



class Libx264_VideoCodecParameters(object):
  def __init__(self):
    self.vprofile = None
    self.level = None
    self.coder = None
    self.flags2 = None
    self.partitions = None
    self.bf = None
    self.directpred = None
    self.me_method = None
    rm_attributes_if_null(self)

  def remove_null_params(self):
    rm_attributes_if_null(self)

class Libx265_VideoCodecParameters(object):
  def __init__(self):
    pass


class MyEncoder(JSONEncoder):
  def default(self, obj):
    return obj.__dict__


class Query(object):
  def __init__(self):
    self.params = None
    self.error = None
    self.message = ''
    self.query = None

  def prepare_params(self):
    query = dict(query=self.params)
    try:
      self.query = json.dumps(query, cls=MyEncoder)
    except BaseException as e:
      self.error = True
      self.message = repr(e)

  def validate_params(self):
    if not self.params:
      self.error = True
      self.message = 'Params is required'
      return
    if not 'source' in self.params.__dict__:
      self.error = True
      self.message = 'Params: source is required'
      return
    if not 'format' in self.params.__dict__:
      self.error = True
      self.message = 'Params: format is required'
      return





"""A setuptools based setup module.
"""

from setuptools import setup
from os import path

here = path.abspath(path.dirname(__file__))


with open('LONG_DESCRIPTION.md') as f:
    long_description = f.read()


setup(
    name='qencode3',
    version='0.9.30',
    description="Qencode client library to easily setup a working solution using Python v3.x.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/qencode-dev/qencode-api-python3-client',
    # url=here,
    author='Qencode Developer',
    author_email='team@qencode.com',
    license='proprietary',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: Other/Proprietary License',
        'Programming Language :: Python :: 3.6'

    ],
    keywords='qencode3, qencode.com, cloud.qencode.com',
    packages=['qencode3']
)   



from . task import *
from qencode3 import QencodeTaskException
import urllib.request as urllib2

class Metadata(Task):

    def get(self, uri):
        params = """
            {"query": {
              "source": "%s",
              "format": [ {"output": "metadata", "metadata_version": "4.1.5"} ]
              }
            }
            """ % uri
        self.custom_start(params)
        while True:
            status = self.status()
            if status['error'] or status['status'] == 'completed':
                break
            time.sleep(5)

        if self.error:
            raise QencodeTaskException(self.message)

        url = None
        if len(status['videos']) > 0:
            url = status['videos'][0]['url']
        elif len(status['audios']) > 0:
            url = status['audios'][0]['url']

        if url is None:
            raise QencodeTaskException('No metadata URL found in status response')

        data = urllib2.urlopen(url).read()

        return data


